package hu.bme.cs.benchmark.mutex.agg;

import agg.xt_basis.Arc;
import agg.xt_basis.BaseFactory;
import agg.xt_basis.GraGra;
import agg.xt_basis.Graph;
import agg.xt_basis.Match;
import agg.xt_basis.Morphism;
import agg.xt_basis.Node;
import agg.xt_basis.Rule;
import agg.xt_basis.Step;
import agg.xt_basis.Type;
import agg.xt_basis.TypeException;

import java.io.File;
// import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Vector;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

public class ALAP {
    private static Logger logger = Logger.getLogger(ALAP.class);
    private static GraGra gt;

    private static Match match(Rule r) {
        Match match = gt.createMatch(r);
        // match.setCompletionStrategy(CompletionStrategySelector.getDefault());
        if(match.nextCompletion()) {
            if(match.isValid()) {
                return match;
            } else {
                return null;
            }
        } else {
            return null;
        }
    }
    
    private static void update(Match match) 
    	throws TypeException {
        Morphism mor = new Step().execute(match);
        /*
        System.out.println("Update");
        Enumeration enum = mor.getOriginal().getNodes();
        while(enum.hasMoreElements()) {
            Node elem = (Node) enum.nextElement();
            System.out.println(elem.getContextUsage() + " -> " + ((Node) mor.getImage(elem)).getContextUsage() + ", " + elem.toString());
        }
        enum = mor.getOriginal().getArcs();
        while(enum.hasMoreElements()) {
            Arc elem = (Arc) enum.nextElement();
            Node src = (Node) elem.getSource();
            Node trg = (Node) elem.getTarget();
            System.out.println(src.getContextUsage() + " -- "
                    + elem.getContextUsage() + " -> "
                    + trg.getContextUsage());
            System.out.println("|");
            System.out.println(((Node) mor.getImage(src)).getContextUsage() + " -- "
                    + ((Arc) mor.getImage(elem)).getContextUsage() + " -> "
                    + ((Node) mor.getImage(trg)).getContextUsage());
        }
        */
    }
    
    private static void apply(Rule r) 
    	throws TypeException {
        if(logger.isDebugEnabled()) {
            logger.debug(r.getName() + ": init: " + System.nanoTime());
            Match match = match(r);
            if(match != null) {
                logger.debug(r.getName() + ": pattern matching: " + System.nanoTime());
                update(match);
                logger.debug(r.getName() + ": update: " + System.nanoTime());
            } else {
                logger.debug(r.getName() + ": no match: " + System.nanoTime());
            }
            gt.destroyMatch(match);
        } else {
            Match match = match(r);
            if(match != null) {
                update(match);
            }
            gt.destroyMatch(match);
        }
    }
    
    public static void showGraph(Graph g) {
        Enumeration enum;
        
        System.out.println(g.getName());
        enum = g.getNodes();
        while(enum.hasMoreElements()) {
            Node elem = (Node) enum.nextElement();
            // System.out.println(elem.getContextUsage() + " : " + elem.getType().getStringRepr());
            System.out.println(elem.toString());
        }
        enum = g.getArcs();
        while(enum.hasMoreElements()) {
            Arc elem = (Arc) enum.nextElement();
            Node src = (Node) elem.getSource();
            Node trg = (Node) elem.getTarget();
            // System.out.print(src.getContextUsage() + " : " + src.getType().getStringRepr());
            System.out.print(src.toString());
            System.out.print(" --- ");
            // System.out.print(elem.getContextUsage() + " : " + elem.getType().getStringRepr());
            System.out.print(elem.toString());
            System.out.print(" --> ");
            // System.out.print(trg.getContextUsage() + " : " + trg.getType().getStringRepr());
            System.out.print(trg.toString());
            System.out.println("");
        }
    }
    
    public static void main(java.lang.String args[]) 
		throws TypeException {

        int N = 0;
        
        try {
            File f = new File(args[1]);
            if(!args[0].equals(new String("-f")) || !f.isFile()) {
                System.err.println("Sample usage: java " + ALAP.class.getName() + " -f ggx_file 4");
                System.exit(1);
            }
            N = Integer.parseInt(args[2]);
        } catch (Exception ex) {
            System.err.println("Sample usage: java " + ALAP.class.getName() + " -f ggx_file 4");
            System.exit(1);
        }
        BasicConfigurator.configure();

        BaseFactory bf = new BaseFactory();
        gt = bf.createGraGra();
        gt.load(args[1]);
        Enumeration enum = gt.getTypes();
        HashMap types = new HashMap(8);
        while(enum.hasMoreElements()) {
            Type elem = (Type) enum.nextElement();
            types.put(elem.getStringRepr(),elem);
        }
    
        Graph g = bf.createGraph(gt.getTypeSet());
        Node p,p1,p2,res;

        /* Initial model generation */
        p1 = g.createNode((Type) types.get("Process"));
        res = g.createNode((Type) types.get("Resource"));
        
        // held_by: r -> p1
        g.createArc((Type) types.get("held_by"),res,p1);
        p = p1;
        int i = 0;
        
        for(i = 1; i < N; i++) {
            p2 = g.createNode((Type) types.get("Process"));
            res = g.createNode((Type) types.get("Resource"));
            // held_by: r -> p2
            g.createArc((Type) types.get("held_by"),res,p2);
            // next: p1 -> p2
            g.createArc((Type) types.get("next"),p1,p2);
            p1 = p2;
        }
        // next: p1 -> p
        g.createArc((Type) types.get("next"),p1,p);
        
        gt = new GraGra(g);
        gt.load(args[1]);
        Vector rules = gt.getRulesVec();
        // showGraph(gt.getGraph());
        
        Rule r;
        for(i = 0; i < N; i++) {
            // release
            r = (Rule) rules.get(7);
            apply(r);
        }
        for(i = 0; i < N; i++) {
            // give
            r = (Rule) rules.get(8);
            apply(r);
        }
        for(i = 0; i < N; i++) {
            // requestSimple
            r = (Rule) rules.get(15);
            apply(r);
        }
        for(i = 0; i < N; i++) {
            // take
            r = (Rule) rules.get(6);
            apply(r);
        }
        // showGraph(gt.getGraph());
    }
}
